# This file is covered by the GPL as part of Rubber.
# (c) Emmanuel Beffara, 2002
"""
Metapost support for Rubber.

The module parses input files for dependencies and does some checkings on
Metapost's log files after the process. Is it enough?
"""

import re

from rubber import _
from rubber.util import *

re_input = re.compile("input +(?P<file>[^ ]+)")
# This is very restrictive, and so is the parsing routine. FIXME?
re_mpext = re.compile("[0-9]+|mpx|log")

class Dep (Depend):
	"""
	This class represents dependency nodes for MetaPost figures. The __init__
	method simply creates one node for the figures and one leaf node for all
	sources.
	"""
	def __init__ (self, target, source, env):
		sources = []
		self.include(source, sources)
		env.msg(2, _("%s is made from %r") % (target, sources))
		self.leaf = DependLeaf(sources)
		Depend.__init__(self, [target], {source: self.leaf})
		self.env = env
		self.base = source[:-3]
		self.cmd = ["mpost", "\\batchmode;input %s" % self.base]
		if env.src_path == "":
			self.penv = {}
		else:
			self.penv = { "MPINPUTS":
				"%s:%s" % (self.env.src_path, os.getenv("MPINPUTS", "")) }

	def include (self, source, list):
		"""
		This function tries to find a specified MetaPost source (currently all
		in the same directory), appends its actual name to the list, and
		parses it to find recursively included files.
		"""
		if exists(source + ".mp"):
			file = source + ".mp"
		elif exists(source):
			file = source
		else:
			return
		list.append(file)
		fd = open(file)
		for line in fd.readlines():
			m = re_input.search(line)
			if m:
				self.include(m.group("file"), list)
		fd.close()

	def run (self):
		self.env.msg(0, _("running Metapost on %s.mp...") % self.base)
		self.env.execute(self.cmd, self.penv)

		# This creates a log file that has the same aspect as TeX logs.

		log = open(self.base + ".log")
		# The semantics of `error' is:
		# 0: no error occured,
		# 1: there has been an error,
		# 2: there has been an error, we are showing it.
		error = 0
		line = log.readline()
		while line != "":
			if error == 2:
				self.env.msg(0, line.rstrip())
				if line[0:2] == "l." or line[0:3] == "***":
					error = 1
			elif line[0] == "!":
				if not error:
					self.env.msg(0, _("There were errors in Metapost code:"))
				self.env.msg(0, line.rstrip())
				error = 2
			line = log.readline()
		log.close()
		return error

	def clean (self):
		"""
		This method removes all the files that the Metapost compilation may
		have created. It is required because the compilation may produce more
		files than just the figures used by the document.
		"""
		base = self.base + "."
		ln = len(base)
		for file in os.listdir("."):
			if file[:ln] == base:
				ext = file[ln:]
				m = re_mpext.match(ext)
				if m and ext[m.end():] == "":
					self.env.msg(3, _("removing %s") % file)
					os.unlink(file)

# The `files' dictionary associates dependency nodes to MetaPost sources. It
# is used to detect when several figures from the same source are included. It
# uses a global variable, and this is authentically BAD. Therefore it deserves
# a big fat huge FIXME. Graphics modules should proably be encapsuled in
# classes the same way as standard Rubber modules, that would help for this
# problem.

files = {}

def convert (source, target, env):
	if files.has_key(source):
		dep = files[source]
		dep.prods.append(target)
	else:
		dep = Dep(target, source, env)
		files[source] = dep
		env.cleaning_process.append(dep.clean)
	return dep
